from io import BytesIO
import cbor2
from cbor2 import CBORTag
# from cbor2 import CBORDecoder
from cbor2.decoder import CBORDecoder

from . import bindings

# ugly patch to remove default UUID decodeur
cbor2.decoder.semantic_decoders.pop(37)


def tag_hook(decoder, tag, shareable_index=None):
    if tag.tag == 37:
        return bindings.UUID(bytes=tag.value)
    if tag.tag == 32:
        return bindings.URL(tag.value)
    return tag

def default_encoder(encoder, value):
    if isinstance(value,bindings.UUID):
        encoder.encode(CBORTag(37, value.bytes))

    if isinstance(value,bindings.URL):
        encoder.encode(CBORTag(32, value.bytes))

def dumps(obj, **kwargs):
    return cbor2.dumps(obj,default=default_encoder,**kwargs)

def loads(payload, **kwargs):
    #return cbor2.loads(payload,tag_hook=tag_hook,**kwargs)
    return _loads(payload,tag_hook=tag_hook,**kwargs)

def _loads(s, **kwargs):
    with BytesIO(s) as fp:
        return CBORDecoder(fp, **kwargs).decode()

#class CustomDecoder(CBORDecoder):pass


def cleanup(obj):
    """ 
    recursive walk a object to search for un-wanted CBOR tags.
    Transform this tag in string format, this can be UUID, URL..
    Should be Ok, with list, dicts.. 
    Warning: This operate in-place changes. 
    Warning: This won't work for tags in dict keys.
    """
    if isinstance(obj,list):
        for i in range(0,len(obj)):
            obj[i] = cleanup(obj[i]) 
        return obj

    if isinstance(obj,dict):
        for k in obj.keys():
            obj.update({k:cleanup(obj[k])})
        return obj

    if type(obj) in bindings.classes:
        return str(obj)
    else:
        return obj



#
#  Copyright 2014, Jérôme Colin, Jérôme Kerdreux, Philippe Tanguy,
#  Telecom Bretagne.
#
#  This file is part of xAAL.
#
#  xAAL is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  xAAL is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with xAAL. If not, see <http://www.gnu.org/licenses/>.
#

from . import core
from .network  import NetworkConnector
from .exceptions  import *
from . import config

import time
import collections
from enum import Enum

import logging
logger = logging.getLogger(__name__)

class EngineState(Enum):
    started  = 1
    running  = 2
    halted   = 3

class Engine(core.EngineMixin):

    __slots__ = ['__last_timer','__txFifo','state','network']

    def __init__(self,address=config.address,port=config.port,hops=config.hops,key=config.key):
        core.EngineMixin.__init__(self,address,port,hops,key)

        self.__last_timer = 0                    # last timer check
        self.__txFifo = collections.deque()      # tx msg fifo

        # message receive workflow
        self.subscribe(self.handle_request)
        # ready to go
        self.state = EngineState.halted
        # start network
        self.network = NetworkConnector(address, port, hops)

    #####################################################
    # xAAL messages Tx handling
    #####################################################
    # Fifo for msg to send
    def queue_msg(self, msg):
        """queue an encoded / cyphered message"""
        self.__txFifo.append(msg)

    def send_msg(self, msg):
        """Send an encoded message to the bus, use queue_msg instead"""
        self.network.send(msg)

    def process_tx_msg(self):
        """ Process (send) message in tx queue called from the loop()"""
        cnt = 0
        while self.__txFifo:
            temp = self.__txFifo.popleft()
            self.send_msg(temp)
            # try to limit rate
            cnt = cnt + 1
            if cnt > config.queue_size:
                time.sleep(0.2)
                break

    #####################################################
    # xAAL messages subscribers
    #####################################################
    def receive_msg(self):
        """return new received message or None"""
        result = None
        data = self.network.get_data()
        if data:
            try:
                msg = self.msg_factory.decode_msg(data,self.msg_filter)
            except MessageParserError as e:
                logger.warning(e)
                msg = None
            result = msg
        return result

    def process_subscribers(self):
        """process incomming messages"""
        msg = self.receive_msg()
        if msg:
            for func in self.subscribers:
                func(msg)
            self.process_attributes_change()
            
    def handle_request(self, msg):
        """
        Filter msg for devices according default xAAL API then process the
        request for each targets identied in the engine
        """
        if not msg.is_request():
            return 
        
        targets = core.filter_msg_for_devices(msg, self.devices)
        for target in targets:
            if msg.action == 'is_alive':
                self.send_alive(target)
            else:
                self.handle_action_request(msg, target)

    def handle_action_request(self, msg, target):
        """
        Run method (xAAL exposed method) on device:
            - None is returned if device method do not return anything
            - result is returned if device method gives a response
            - Errors are raised if an error occured:
                * Internal error
                * error returned on the xAAL bus
        """
        try:
            result = run_action(msg, target)
            if result != None:
                self.send_reply(dev=target,targets=[msg.source],action=msg.action,body=result)
        except CallbackError as e:
            self.send_error(target, e.code, e.description)
        except XAALError as e:
            logger.error(e)

    #####################################################
    # timers
    #####################################################
    def process_timers(self):
        """Process all timers to find out which ones should be run"""
        expire_list = []
        
        if len(self.timers)!=0 :
            now = time.time()
            # little hack to avoid to check timer to often.
            # w/ this enable timer precision is bad, but far enougth
            if (now - self.__last_timer) < 0.4: return

            for t in self.timers:
                if t.deadline < now:
                    try:
                        t.func()
                    except CallbackError as e:
                        logger.error(e.description)
                    if t.counter != -1:
                        t.counter-= 1
                        if t.counter == 0:
                            expire_list.append(t)
                    t.deadline = now + t.period
            # delete expired timers
            for t in expire_list:
                self.remove_timer(t)
                    
            self.__last_timer = now

    #####################################################
    # Mainloops & run ..
    #####################################################
    def loop(self):
        """
        Process incomming xAAL msg
        Process timers
        Process attributes change for devices
        Process is_alive for device
        Send msgs from the Tx Buffer
        """
        # Process xAAL msg received, filter msg and process request
        self.process_subscribers()
        # Process timers
        self.process_timers()
        # Process attributes change for devices due to timers
        self.process_attributes_change()
        # Process Alives
        self.process_alives()
        # Process xAAL msgs to send
        self.process_tx_msg()

    def start(self):
        """Start the core engine: send queue alive msg"""
        if self.state in [EngineState.started,EngineState.running]:
            return
        self.network.connect()
        for dev in self.devices:
            self.send_alive(dev)
        self.state = EngineState.started

    def stop(self):
        self.state = EngineState.halted

    def shutdown(self):
        self.stop()

    def run(self):
        self.start()
        self.state = EngineState.running
        while self.state == EngineState.running:
            self.loop()

    def is_running(self):
        if self.state == EngineState.running:
            return True
        return False

def run_action(msg,device):
    """ 
    Extract an action & launch it
    Return:
        - action result
        - None if no result

    Note: If an exception raised, it's logged, and raise an XAALError.
    """
    method,params = core.search_action(msg,device)
    result = None
    try:
        result = method(**params)
    except Exception as e:
        logger.error(e)
        raise XAALError("Error in method:%s params:%s" % (msg.action,params))
    return result




#
#  Copyright 2014, Jérôme Colin, Jérôme Kerdreux, Philippe Tanguy,
#  Telecom Bretagne.
#
#  This file is part of xAAL.
#
#  xAAL is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  xAAL is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with xAAL. If not, see <http://www.gnu.org/licenses/>.
#


from . import config
from . import tools
from . import bindings
from .exceptions import DeviceError

from tabulate import tabulate
import logging
logger = logging.getLogger(__name__)

import time

class Attribute(object):

    def __init__(self, name, dev=None, default=None):
        self.name = name
        self.default = default
        self.device = dev
        self.__value = default

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, value):
        if value != self.__value:
            eng = self.device.engine
            if eng:
                eng.add_attributes_change(self)
                logger.debug("Attr change %s %s=%s" % (self.device.address,self.name,value))
        self.__value = value

    def __repr__(self): # pragma: no cover
        return f"<{self.__module__}.Attribute {self.name} at 0x{id(self):x}>"
    

class Attributes(list):
    """Devices owns a attributes list. This list also have dict-like access"""

    def __getitem__(self,value):
        if isinstance(value,int):
            return list.__getitem__(self,value)
        for k in self:
            if (value == k.name):
                return k.value
        raise KeyError(value)
    
    def __setitem__(self,name,value):
        if isinstance(name,int):
            return list.__setitem__(self,name,value)
        for k in self:
            if (name == k.name):
                k.value = value
                return
        raise KeyError(name)
    
class Device(object):

    __slots__ = ['__dev_type','__address','group_id',
                'vendor_id','product_id','hw_id',
                '__version','__url','schema','info',
                'unsupported_attributes','unsupported_methods','unsupported_notifications',
                'alive_period','next_alive',
                '__attributes','methods','engine']

    def __init__(self,dev_type,addr=None,engine=None):
        # xAAL internal attributes for a device
        self.dev_type = dev_type        # xaal dev_type
        self.address = addr             # xaal addr
        self.group_id = None            # group devices
        self.vendor_id = None           # vendor ID ie : ACME
        self.product_id = None          # product ID
        self.hw_id = None               # hardware info
        self.__version = None           # product release
        self.__url = None               # product URL
        self.schema = None              # schema URL
        self.info = None                # additionnal info
        
        # Unsupported stuffs
        self.unsupported_attributes = []
        self.unsupported_methods = []
        self.unsupported_notifications = []
        # Alive management
        self.alive_period = config.alive_timer # time in sec between two alive
        self.next_alive = 0
        # Default attributes & methods
        self.__attributes = Attributes()
        self.methods = {'get_attributes' : self._get_attributes,
                        'get_description': self._get_description }
        self.engine = engine

    @property
    def dev_type(self):
        return self.__dev_type

    @dev_type.setter
    def dev_type(self, value):
        if not tools.is_valid_dev_type(value):
            raise DeviceError(f"The dev_type {value} is not valid")
        self.__dev_type = value

    @property
    def version(self):
        return self.__version

    @version.setter
    def version(self, value):
        # version must be a string
        if value:
            self.__version = "%s" % value
        else:
            self.__version = None

    @property
    def address(self):
        return self.__address

    @address.setter
    def address(self, value):
        if value == None:
            self.__address = None
            return
        if not tools.is_valid_address(value):
            raise DeviceError("This address is not valid")
        self.__address = value

    @property
    def url(self):
        return self.__url

    @url.setter
    def url(self,value):
        if value == None:
            self.__url = None
        else:
            self.__url = bindings.URL(value)

    # attributes
    def new_attribute(self,name,default=None):
        attr = Attribute(name,self,default)
        self.add_attribute(attr)
        return attr

    def add_attribute(self, attr):
        if attr:
            self.__attributes.append(attr)
            attr.device = self
    
    def del_attribute(self,attr):
        if attr:
            attr.device = None
            self.__attributes.remove(attr)

    def get_attribute(self,name):
        for attr in self.__attributes:
            if attr.name == name:
                return attr
        return None

    @property
    def attributes(self):
        return self.__attributes

    @attributes.setter
    def attributes(self,values):
        if isinstance(values,Attributes):
            self.__attributes = values
        else:
            raise DeviceError("Invalid attributes list, use class Attributes)")

    def add_method(self,name,func):
        self.methods.update({name:func})

    def get_methods(self):
        return self.methods

    def update_alive(self):
        """ update the alive timimg"""
        self.next_alive = time.time() + self.alive_period

    def get_timeout(self):
        """ return Alive timeout used for isAlive msg"""
        return 2 * self.alive_period

    #####################################################
    # Usefull methods
    #####################################################
    def dump(self):
        print("= Device: %s" % self)
        # info & description
        r = []
        r.append(['dev_type',self.dev_type])
        r.append(['address',self.address])
        for k,v in self._get_description().items():
            r.append([k,v])
        print(tabulate(r,tablefmt="fancy_grid"))
        
        # attributes
        if len(self._get_attributes()) > 0:
            r = []
            for k,v in self._get_attributes().items():
                r.append([k,str(v)])
            print(tabulate(r,tablefmt="fancy_grid"))

        # methods
        if len(self.methods) > 0:
            r = []
            for k,v in self.methods.items():
                r.append([k,v.__name__])
            print(tabulate(r,tablefmt="fancy_grid"))


    def __repr__(self):
        return f"<xaal.Device {id(self):x} {self.address} {self.dev_type}>"

    #####################################################
    # default public methods
    #####################################################
    def _get_description(self):
        result = {}
        if self.vendor_id:  result['vendor_id']  = self.vendor_id
        if self.product_id: result['product_id'] = self.product_id
        if self.version:    result['version']    = self.version
        if self.url:        result['url']        = self.url
        if self.schema:     result['schema']     = self.schema
        if self.info:       result['info']       = self.info
        if self.hw_id:      result['hw_id']      = self.hw_id
        if self.group_id:   result['group_id']   = self.group_id
        if self.unsupported_methods:       result['unsupported_methods']       = self.unsupported_methods
        if self.unsupported_notifications: result['unsupported_notifications'] = self.unsupported_notifications
        if self.unsupported_attributes:    result['unsupported_attributes']    = self.unsupported_attributes
        return result

    def _get_attributes(self, _attributes=None):
        """
        attributes:
            - None = body empty and means request all attributes
            - Empty array means request all attributes
            - Array of attributes (string) and means request attributes in the
              list

        TODO: (Waiting for spec. decision) add test on attribute devices
            - case physical sensor not responding or value not ready add error
            with specific error code and with value = suspicious/stale/cached
        """
        result = {}
        dev_attr = {attr.name: attr for attr in self.__attributes}
        if _attributes:
            # Process attributes filter
            for attr in _attributes:
                if attr in dev_attr.keys():
                    result.update({dev_attr[attr].name: dev_attr[attr].value})
                else:
                    logger.debug(f"Attribute {attr} not found")
        else:
            # Process all attributes
            for attr in dev_attr.values():
                result.update({attr.name: attr.value})
        return result

    def send_notification(self,notification,body={}):
        """ queue an notification, this is just a method helper """
        if self.engine:
            self.engine.send_notification(self,notification,body)



import asyncio

from . import core
from . import config
from . import tools
from .messages import MessageParserError
from .aionetwork import AsyncNetworkConnector
from .exceptions import *

import time
from enum import Enum
import aioconsole
import signal
import sys
from tabulate import tabulate
from pprint import pprint


import logging
logger = logging.getLogger(__name__)

class AsyncEngine(core.EngineMixin):

    __slots__ = ['__txFifo','_loop','_tasks','_hooks','_watchdog_task','_kill_counter','running_event','watchdog_event','started_event']

    def __init__(self,address=config.address,port=config.port,hops=config.hops,key=config.key):
        core.EngineMixin.__init__(self,address,port,hops,key)

        self.__txFifo = asyncio.Queue()          # tx msg fifo
        self._loop = None                        # event loop
        self._hooks = []                         # hooks
        self._tasks = []                         # tasks
        self._watchdog_task = None               # watchdog task
        self._kill_counter = 0                   # watchdog counter

        self.started_event  = asyncio.Event()   # engine started event
        self.running_event  = asyncio.Event()   # engine running event
        self.watchdog_event = asyncio.Event()   # watchdog event

        signal.signal(signal.SIGTERM, self.sigkill_handler)
        signal.signal(signal.SIGINT, self.sigkill_handler)
        
        # message receive workflow
        self.subscribe(self.handle_request)
        # start network
        self.network = AsyncNetworkConnector(address, port, hops)

    #####################################################
    # Hooks
    #####################################################
    def on_start(self,func,*args,**kwargs):
        hook = Hook(HookType.start,func,*args,**kwargs)
        self._hooks.append(hook)
    
    def on_stop(self,func,*args,**kwargs):
        hook = Hook(HookType.stop,func,*args,**kwargs)
        self._hooks.append(hook)

    async def run_hooks(self,hook_type):
        hooks = list(filter(lambda hook: hook.type==hook_type,self._hooks))
        if len(hooks)!=0:
            logger.debug(f"Launching {hook_type} hooks")
            for h in hooks:
                await run_func(h.func,*h.args,**h.kwargs)

    #####################################################
    # timers
    #####################################################
    async def process_timers(self):
        """Process all timers to find out which ones should be run"""
        expire_list = []
        now = time.time()
        for t in self.timers:
            if t.deadline < now:
                try:
                    await run_func(t.func)
                except CallbackError as e:
                    logger.error(e.description)
                if t.counter != -1:
                    t.counter-= 1
                    if t.counter == 0:
                        expire_list.append(t)
                t.deadline = now + t.period
        # delete expired timers
        for t in expire_list:
            self.remove_timer(t)

    #####################################################
    # msg send / receive
    #####################################################
    def queue_msg(self, msg):
        """queue a message"""
        self.__txFifo.put_nowait(msg)

    def send_msg(self, msg):
        """Send an encoded message to the bus, use queue_msg instead"""
        self.network.send(msg)

    async def receive_msg(self):
        """return new received message or None"""
        data = await self.network.get_data()
        if data:
            try:
                msg = self.msg_factory.decode_msg(data,self.msg_filter)
            except MessageParserError as e:
                logger.warning(e)
                msg = None
            return msg
        return None

    async def process_rx_msg(self):
        """process incomming messages"""
        msg = await self.receive_msg()
        if msg:
            for func in self.subscribers:
                await run_func(func,msg)
            self.process_attributes_change()

    def handle_request(self, msg):
        """Filter msg for devices according default xAAL API then process the
        request for each targets identied in the engine
        """
        if not msg.is_request():
            return
        targets = core.filter_msg_for_devices(msg, self.devices)
        for target in targets:
            if msg.action == 'is_alive':
                self.send_alive(target)
            else:
                self.new_task(self.handle_action_request(msg, target))

    async def handle_action_request(self, msg, target):
        try:
            result = await run_action(msg, target)
            if result != None:
                self.send_reply(dev=target,targets=[msg.source],action=msg.action,body=result)
        except CallbackError as e:
            self.send_error(target, e.code, e.description)
        except XAALError as e:
            logger.error(e)

    #####################################################
    # Asyncio loop & Tasks
    #####################################################
    def get_loop(self):
        if self._loop == None:
            logger.debug('New event loop')
            self._loop = asyncio.get_event_loop()
        return self._loop

    def new_task(self,coro,name=None):
        # we maintain a task list, to be able to stop/start the engine
        # on demand. needed by HASS
        task = self.get_loop().create_task(coro,name=name)
        self._tasks.append(task)
        task.add_done_callback(self.task_callback)
        return task

    def task_callback(self, task):
        # called when a task ended
        self._tasks.remove(task)

    def all_tasks(self):
        return self._tasks

    async def boot_task(self):
        self.watchdog_event.clear()
        # queue the alive before anything
        for dev in self.devices:
            self.send_alive(dev)
        await self.network.connect()
        self.running_event.set()
        await self.run_hooks(HookType.start)

    async def receive_task(self):
        await self.running_event.wait()
        while self.is_running():
            await self.process_rx_msg()

    async def send_task(self):
        await self.running_event.wait()
        while self.is_running():
            temp = await self.__txFifo.get()
            self.send_msg(temp)

    async def timer_task(self):
        await self.running_event.wait()
        self.setup_alives_timer()
        while self.is_running():
            await asyncio.sleep(0.2)
            await self.process_timers()
            self.process_attributes_change()

    async def watchdog_task(self):
        await self.watchdog_event.wait()
        await self.stop()
        logger.info('Exit')
    
    #####################################################
    # start / stop / shutdown
    #####################################################
    def is_running(self):
        return self.running_event.is_set()

    def start(self):
        if self.is_running():
            logger.warning('Engine already started')
            return
        self.started_event.set()
        self.new_task(self.boot_task(),name='Boot')
        self.new_task(self.receive_task(),name='RecvQ')
        self.new_task(self.send_task(),name='SendQ')
        self.new_task(self.timer_task(),name='Timers')
        self.new_task(console(locals()),name='Console')

    def setup_alives_timer(self):
        # needed on stop-start sequence
        if self.process_alives in [t.func for t in self.timers]:
            return
        # process alives every 10 seconds
        self.add_timer(self.process_alives,10)

    async def stop(self):
        logger.info('Stopping engine')
        await self.run_hooks(HookType.stop)
        self.running_event.clear()
        self.started_event.clear()
        # cancel all tasks
        for task in self.all_tasks():
            if task!=self._watchdog_task:
                task.cancel()
                await asyncio.sleep(0.1)

    def sigkill_handler(self,signal,frame):
        print("", end = "\r") #remove the uggly ^C
        if not self.is_running():
            logger.warning('Engine already stopped')
            self._kill_counter = 1
        self._kill_counter +=1
        self.shutdown()
        if self._kill_counter > 1:
            logger.warning('Force quit')
            sys.exit(-1)
        else:
            logger.warning('Kill requested')

    def shutdown(self):
        self.watchdog_event.set()
    
    def run(self):
        if not self.started_event.is_set():
            self.start()
        if self._watchdog_task == None:
            # start the watchdog task
            self._watchdog_task = self.new_task(self.watchdog_task(),name='Watchdog task')
            self.get_loop().run_until_complete(self._watchdog_task)
        else:
            logger.warning('Engine already running')

    #####################################################
    # Debugging tools
    #####################################################
    def dump_timers(self):
        headers = ['Func','Period','Counter','Deadline']
        rows = []
        now = time.time()
        for t in self.timers:
            remain = round(t.deadline-now,1)
            rows.append([str(t.func),t.period,t.counter,remain])
        print('= Timers')
        print(tabulate(rows,headers=headers,tablefmt="fancy_grid"))


    def dump_tasks(self):
        headers = ["Name","Coro","Loop ID"]
        rows = []
        for t in self.all_tasks():
            rows.append([t.get_name(),str(t.get_coro()),id(t.get_loop())])
        print('= Tasks')
        print(tabulate(rows,headers=headers,tablefmt="fancy_grid"))

    def dump_devices(self):
        headers = ["addr","dev_type","info"]
        rows = []
        for d in self.devices:
            rows.append([d.address,d.dev_type,d.info])
        print('= Devices')
        print(tabulate(rows,headers=headers,tablefmt="fancy_grid"))

    def dump_hooks(self):
        headers = ["Type","Hook"]
        rows = []
        for h in self._hooks:
            rows.append([h.type,str(h.func)])
        print('= Hooks')
        print(tabulate(rows,headers=headers,tablefmt="fancy_grid"))

    def dump(self):
        self.dump_devices()
        self.dump_tasks()
        self.dump_timers()
        self.dump_hooks()

    def get_device(self,uuid):
        uuid = tools.get_uuid(uuid)        
        for dev in self.devices:
            if dev.address == uuid:
                return dev
        return None


#####################################################
# Utilities functions
#####################################################
async def run_func(func,*args,**kwargs):
    """run a function or a coroutine function """
    if asyncio.iscoroutinefunction(func):
        return await func(*args,**kwargs)
    else:
        return func(*args,**kwargs)


async def run_action(msg,device):
    """ 
    Extract an action & launch it
    Return:
        - action result
        - None if no result

    Notes: 
        - If an exception raised, it's logged, and raise an XAALError.
        - Same API as legacy Engine, but accept coroutine functions
    """
    method,params = core.search_action(msg,device)
    result = None
    try:
        if asyncio.iscoroutinefunction(method):
            result = await method(**params)
        else:
            result = method(**params)
    except Exception as e:
        logger.error(e)
        raise XAALError("Error in method:%s params:%s" % (msg.action,params))
    return result


#####################################################
# Hooks
#####################################################
class HookType(Enum):
    start = 0
    stop  = 1

class Hook(object):
    __slots__ = ['type','func','args','kwargs']
    def __init__(self,type_,func,*args,**kwargs):
        self.type = type_
        self.func = func
        self.args = args
        self.kwargs = kwargs


#####################################################
# Debugging console
#####################################################
async def console(locals=locals(),port=None):
    """launch a console to enable remote engine inspection"""
    if port == None:
        # let's find a free port if not specified
        def find_free_port():
            import socketserver
            with socketserver.TCPServer(("localhost", 0), None) as s:
                return s.server_address[1]
        port = find_free_port()
    
    logger.debug(f'starting debug console on port {port}')
    sys.ps1 = '[xAAL] >>> '
    banner = '=' * 78 +"\nxAAL remote console\n" + '=' *78
    locals.update({'pprint':pprint})

    def factory(streams):
        return aioconsole.AsynchronousConsole(locals=locals, streams=streams)
    # start the console
    try:
        # debian with ipv6 disabled still state that localhost is ::1, which broke aioconsole
        await aioconsole.start_interactive_server(host='127.0.0.1', port=port,factory=factory,banner=banner)
    except OSError:
        logger.warning('Unable to run console')



#
#  Copyright 2014, Jérôme Colin, Jérôme Kerdreux, Philippe Tanguy,
#  Telecom Bretagne.
#
#  This file is part of xAAL.
#
#  xAAL is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  xAAL is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with xAAL. If not, see <http://www.gnu.org/licenses/>.
#

import os
import re

import pysodium

import sys
import functools
from configobj import ConfigObj

from . import config
from .bindings import UUID

XAAL_DEVTYPE_PATTERN = '^[a-zA-Z][a-zA-Z0-9_-]*\.[a-zA-Z][a-zA-Z0-9_-]*$'

def get_cfg_filename(name, cfg_dir=config.conf_dir):
    if name.startswith('xaal.'):
        name = name[5:]
    filename = '%s.ini' % name
    if not os.path.isdir(cfg_dir):
        print("Your configuration directory doesn't exist: [%s]" % cfg_dir)
    return os.path.join(cfg_dir, filename)
        
def load_cfg_file(filename):
    """ load .ini file and return it as dict"""
    if os.path.isfile(filename):
        return ConfigObj(filename,indent_type='  ',encoding="utf8")
    return None

def load_cfg(app_name):
    filename = get_cfg_filename(app_name)
    return load_cfg_file(filename)

def load_cfg_or_die(app_name):
    cfg = load_cfg(app_name)
    if not cfg:
        print("Unable to load config file %s" % get_cfg_filename(app_name))
        sys.exit(-1)
    return cfg

def new_cfg(app_name):
    filename = get_cfg_filename(app_name)
    cfg = ConfigObj(filename,indent_type='  ')
    cfg['config'] = {}
    cfg['config']['addr']=get_random_uuid().str
    return cfg

def get_random_uuid():
    return UUID.random()

def get_random_base_uuid(digit=2):
    return UUID.random_base(digit)

def get_uuid(val):
    if isinstance(val,UUID):
        return val
    if isinstance(val,str):
        return str_to_uuid(val)
    return None

def str_to_uuid(val):
    """ return an xAAL address for a given string"""
    try:
        return UUID(val)
    except ValueError:
        return None

def bytes_to_uuid(val):
    try:
        return UUID(bytes=val)
    except ValueError:
        return None

def is_valid_uuid(val):
    return isinstance(val,UUID)

def is_valid_address(val):
    return is_valid_uuid(val)

@functools.lru_cache(maxsize=128)
def is_valid_dev_type(val):
    if not isinstance(val,str):
        return False
    if re.match(XAAL_DEVTYPE_PATTERN,val):
       return True
    return False

def pass2key(passphrase):
    """Generate key from passphrase using libsodium
    crypto_pwhash_scryptsalsa208sha256 func
    salt: buffer of zeros
    opslimit: crypto_pwhash_scryptsalsa208sha256_OPSLIMIT_INTERACTIVE
    memlimit: crypto_pwhash_scryptsalsa208sha256_MEMLIMIT_INTERACTIVE
    """
    buf = passphrase.encode('utf-8')
    KEY_BYTES = pysodium.crypto_pwhash_scryptsalsa208sha256_SALTBYTES #32 
    # this should be:
    # salt = bytes(KEY_BYTES)
    # but due to bytes() stupid stuff in py2 we need this awfull stuff
    salt = ('\00' * KEY_BYTES).encode('utf-8')
    opslimit = pysodium.crypto_pwhash_scryptsalsa208sha256_OPSLIMIT_INTERACTIVE
    memlimit = pysodium.crypto_pwhash_scryptsalsa208sha256_MEMLIMIT_INTERACTIVE
    key = pysodium.crypto_pwhash_scryptsalsa208sha256(KEY_BYTES, buf, salt, opslimit, memlimit)
    return key

@functools.lru_cache(maxsize=128)
def reduce_addr(addr):
    """return a string based addred without all  digits"""
    tmp = addr.str
    return tmp[:5] + '..' + tmp[-5:]





# Default configuration
import os
import sys
import binascii
from configobj import ConfigObj

self = sys.modules[__name__]

# Default settings
DEF_ADDR = '224.0.29.200'       # mcast address
DEF_PORT = 1236                 # mcast port
DEF_HOPS = 10                   # mcast hop
DEF_ALIVE_TIMER = 100           # Time between two alive msg
DEF_CIPHER_WINDOW = 60 * 2      # Time Window in seconds to avoid replay attacks
DEF_QUEUE_SIZE = 10             # How many packet we can send in one loop
DEF_LOG_LEVEL = 'DEBUG'         # should be INFO|DEBUG|None
DEF_LOG_PATH  = '/var/log/xaal' # where log are 

# TBD : Move this stuff
STACK_VERSION = 7


if 'XAAL_CONF_DIR' in os.environ:
    self.conf_dir = os.environ['XAAL_CONF_DIR']
else:
    self.conf_dir = os.path.expanduser("~") + '/.xaal'


def load_config(name='xaal.ini'):
    filename = os.path.join(self.conf_dir, name)
    if not os.path.isfile(filename):
        print("Unable to load xAAL config file [%s]" % filename)
        sys.exit(-1)

    cfg = ConfigObj(filename)
    self.address       = cfg.get('address',DEF_ADDR)
    self.port          = int(cfg.get('port',DEF_PORT))
    self.hops          = int(cfg.get('hops',DEF_HOPS))
    self.alive_timer   = int(cfg.get('alive_timer',DEF_ALIVE_TIMER))
    self.cipher_window = int(cfg.get('ciper_window',DEF_CIPHER_WINDOW))
    self.queue_size    = int(cfg.get('queue_size',DEF_QUEUE_SIZE))
    self.log_level     = cfg.get('log_level',DEF_LOG_LEVEL)
    self.log_path      = cfg.get('log_path',DEF_LOG_PATH)
    key = cfg.get('key',None)

    if key:
        self.key = binascii.unhexlify(key.encode('utf-8'))
    else:
        print("Please set key in config file [%s]" % filename)
        self.key = None

load_config()



import asyncio
import struct
import socket

import logging

logger = logging.getLogger(__name__)

class AsyncNetworkConnector(object):

    def __init__(self, addr, port, hops,bind_addr='0.0.0.0'):
        self.addr = addr
        self.port = port
        self.hops = hops
        self.bind_addr = bind_addr
        self._rx_queue = asyncio.Queue()

    async def connect(self):
        loop = asyncio.get_running_loop()
        on_con_lost = loop.create_future()
        self.transport, self.protocol = await loop.create_datagram_endpoint( 
            lambda: XAALServerProtocol(on_con_lost,self.receive), sock = self.new_sock())
        # In some conditions (containers), transport is connected but IGMP is delayed (up to 10ms)
        # so we need to wait for IGMP to be really sent.
        await asyncio.sleep(0.05)

    def new_sock(self):
        sock = socket.socket(socket.AF_INET,socket.SOCK_DGRAM,socket.IPPROTO_UDP)
        try:
            # Linux + MacOS + BSD
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        except:
            # Windows
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((self.bind_addr, self.port))
        mreq = struct.pack("=4s4s",socket.inet_aton(self.addr),socket.inet_aton(self.bind_addr))
        sock.setsockopt(socket.IPPROTO_IP,socket.IP_ADD_MEMBERSHIP,mreq)
        sock.setsockopt(socket.IPPROTO_IP,socket.IP_MULTICAST_TTL,10)
        sock.setblocking(False)
        return sock

    def send(self,data):
        self.protocol.datagram_send(data,self.addr,self.port)

    def receive(self,data):
        self._rx_queue.put_nowait(data)

    async def get_data(self):
        return await self._rx_queue.get()

class XAALServerProtocol(asyncio.Protocol):
    def __init__(self,on_con_lost,on_dtg_recv):
        self.on_con_lost = on_con_lost
        self.on_dtg_recv = on_dtg_recv
        
    def connection_made(self, transport):
        logger.info("xAAL network connected")
        self.transport = transport

    def error_received(self, exc):
        print('Error received:', exc)
        logger.warning(f"Error received: {exc}")

    def connection_lost(self, exc):
        logger.info(f"Connexion closed: {exc}")
        self.on_con_lost.set_result(True)
    
    def datagram_send(self,data,ip,port):
        self.transport.sendto(data,(ip,port))

    def datagram_received(self, data, addr):
        #print(f"pkt from {addr}")
        self.on_dtg_recv(data)





# devices.py
class DeviceError(Exception):pass

# core.py
class EngineError(Exception):pass
class XAALError(Exception):pass
class CallbackError(Exception):
    def __init__(self, code, desc):
        self.code = code
        self.description = desc

# messages.py
class MessageParserError(Exception):pass
class MessageError(Exception):pass

# binding.py
class UUIDError(Exception):pass

__all__ = ["DeviceError","EngineError","XAALError","CallbackError","MessageParserError","MessageError","UUIDError"]



import uuid

from .exceptions import UUIDError

class UUID:
    def __init__(self,*args,**kwargs):
        self.__uuid = uuid.UUID(*args,**kwargs)

    @staticmethod
    def random_base(digit=2):
        """zeros the last digits of a random uuid, usefull w/ you want to forge some addresses
            two digit is great.
        """
        if (digit > 0)  and  (digit < 13): 
            tmp = str(uuid.uuid1())
            st = "%s%s" % (tmp[:-digit],'0'*digit)
            return UUID(st)
        else:
            raise UUIDError

    @staticmethod
    def random():
        tmp = uuid.uuid1().int
        return UUID(int=tmp)

    def __add__(self,value):
        tmp = self.__uuid.int + value
        return UUID(int=tmp)

    def __sub__(self,value):
        tmp = self.__uuid.int - value
        return UUID(int=tmp)

    def __eq__(self,value):
        return self.__uuid == value

    def __lt__(self, value ):
        return self.__uuid.int < value

    def __gt__(self, value ):
        return self.__uuid.int > value

    def __str__(self):
        return str(self.__uuid)
    
    def __repr__(self): # pragma: no cover
        return f"UUID('{self.__uuid}')"

    def __hash__(self):
        return self.__uuid.__hash__()

    def get(self):
        return self.__uuid
    
    def set(self,value):
        self.__uuid = value

    @property
    def str(self):
        return str(self)

    @property
    def bytes(self):
        return self.__uuid.bytes



class URL:
    def __init__(self,value):
        self.__url = value

    def __eq__(self,value):
        return self.__url == value

    def __str__(self):
        return str(self.__url)
    
    def __repr__(self): # pragma: no cover
        return f"URL('{self.__url}')" 

    def set(self,value):
        self.__url = value

    def get(self):
        return self.__url

    @property
    def str(self):
        return str(self)

    @property
    def bytes(self):
        return self.__url



classes = [UUID,URL]




#
#  Copyright 2014, Jérôme Colin, Jérôme Kerdreux, Philippe Tanguy,
#  Telecom Bretagne.
#
#  This file is part of xAAL.
#
#  xAAL is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  xAAL is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with xAAL. If not, see <http://www.gnu.org/licenses/>.
#

import socket
import struct
import select
import logging

logger = logging.getLogger(__name__)

import time
from enum import Enum

class NetworkState(Enum):
    disconnected = 0
    connected    = 1


class NetworkConnector(object):
    UDP_MAX_SIZE = 65507

    def __init__(self, addr, port, hops,bind_addr='0.0.0.0'):
        self.addr = addr
        self.port = port
        self.hops = hops
        self.bind_addr = bind_addr
        self.state = NetworkState.disconnected

    def connect(self):
        try:
            self.__connect()
        except Exception as e:
            self.network_error(e)

    def __connect(self):
        logger.info("Connecting to %s:%s" % (self.addr, self.port))

        self.__sock = socket.socket(socket.AF_INET,socket.SOCK_DGRAM,socket.IPPROTO_UDP)
        self.__sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        # #formac os ???
        #self.__sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.__sock.bind((self.bind_addr, self.port))
        mreq = struct.pack("=4s4s",socket.inet_aton(self.addr),socket.inet_aton(self.bind_addr))
        self.__sock.setsockopt(socket.IPPROTO_IP,socket.IP_ADD_MEMBERSHIP,mreq)
        self.__sock.setsockopt(socket.IPPROTO_IP,socket.IP_MULTICAST_TTL,self.hops)
        self.state = NetworkState.connected

    def disconnect(self):
        logger.info("Disconnecting from the bus")
        self.state = NetworkState.disconnected
        self.__sock.close()

    def is_connected(self):
        return self.state == NetworkState.connected

    def receive(self):
        packt = self.__sock.recv(self.UDP_MAX_SIZE)
        return packt

    def __get_data(self):
        r = select.select([self.__sock, ], [], [], 0.02)
        if r[0]:
            return self.receive()
        return None

    def get_data(self):
        if not self.is_connected(): self.connect()
        try:
            return self.__get_data()
        except Exception as e:
            self.network_error(e)

    def send(self,data):
        if not self.is_connected(): self.connect()
        try:
            self.__sock.sendto(data, (self.addr, self.port))
        except Exception as e:
            self.network_error(e)

    def network_error(self, msg):
        self.disconnect()
        logger.info("Network error, reconnect..%s" % msg)
        time.sleep(5)
